#version 150
precision highp float;

//Ethan Alexander Shulman "Cloud" January 7, 2021

in vec3 in_Position;
out float time;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
#define screen vec2(screenX,screenY)

#define ANIMATION 1
#define FSAMPLES 1000
//#define DEBUG 1
//#define HAS_OPACITY 1


#define STEPS 30
#define EPS 1e-2
#define REFRACTION_SKIP 5.
#define PI 3.14159265

mat2 r2(float a) {
	float s = sin(a), c = cos(a);
	return mat2(c,-s,s,c);
}

mat2x3 tangentBinormal(vec3 norm) {
	mat2x3 t;
	if (abs(norm.x) > abs(norm.y)) t[0] = normalize(vec3(-norm.z,0,norm.x));
	else t[0] = normalize(vec3(0,norm.z,-norm.y)); 
	t[1] = cross(norm,t[0]);
	return t;
}


vec3 hash(uvec3 x) {
    x = ((x>>8U)^x.yzx)*1103515245U;
    x = ((x>>8U)^x.yzx)*1103515245U;
    x = ((x>>8U)^x.yzx)*1103515245U;
    
    return vec3(x)*(1.0/float(0xffffffffU));
}

vec3 hsvToRGB(vec3 hsv) {
	return mix(clamp(abs(mod(hsv.x*6.0+vec3(0.,4.,2.),6.)-3.0)-1.0,0.,1.),vec3(1),hsv.y)*hsv.z;
}

float twave(float x) {
	return 1.-abs(fract(x)-.5)*2.;
}

float nsampling;

//volumetric opacity via distance dithering
#define START_VOLUME(posSeed) vec3 opacityRand = hash(uvec3(abs(posSeed)*1000.)+uvec3(frame,frame*3,frame*7))
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip*(.5+opacityRand.y)+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

float geometry(vec3 p) {
	float d = abs(length(p)-2.)-.7;//sphere base shape
	
	//cloud distortion
	d += twave(length(p-vec3(-4.81, 9.58, 5.85))*1.21)*.24;
	d += twave(length(p-vec3(-8.82, -1.95, 6.94))*.98)*.21;
	d += twave(length(p-vec3(2.69, 3.32, 7.48))*4.98)*.1;
	d += twave(length(p-vec3(6.71, -6.07, -0.94))*6.36)*.08;
	d += twave(length(p-vec3(-3.42, 7.45, 9.32))*5.042)*.1;
	
	//volume dither
	START_VOLUME(p);
	return VOLUME(d, .03, .1, 1.);
}

//chromatic separation dither
#define START_CHROMA(posSeed) float chromaFrame = float(fract(hash(uvec3(abs(posSeed)*1000.)+uvec3(frame,frame*3,frame*7)).x)*3.)
#define CHROMA(mat,rc,gc,bc) if (chromaFrame < 1.) {mat[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mat[0]*=vec3(0,3,0);gc;}\
else {mat[0]*=vec3(0,0,3);bc;}

/*mat[0] = color
mat[1].x = forward roughness, mat[1].y = emission, mat[1].z = backward roughness
mat[2].x = opacity, mat[2].y = refraction index
*/
mat3 material(vec3 p) {
	mat3 m = mat3(vec3(.8), vec3(.1,0.,1.), vec3(1.,1.,0.));//1.517,0.));
	return m;
}
vec3 normal(vec3 p) {
	#define NEPS vec3(EPS,0,0)
	nsampling = 1.;
	vec3 n = normalize(vec3(geometry(p+NEPS),geometry(p+NEPS.yxz),geometry(p+NEPS.yzx))-geometry(p));
	nsampling = 0.;
	return n!=n?vec3(0,1,0):n;
}

vec3 background(vec3 rp, vec3 rd) {
	return mix(vec3(.00013,.00014,.00012),vec3(.001,.002,.006),clamp(rd.y*10.,0.,1.))+
			vec3(1.,.85,.6)*10.*max(0.,1.-length(rd-vec3(0,.3,.2)))*max(0.,rd.y);
}


uniform sampler2D Buffer0;

void vert() {
	time = float(frame)*.1/float(framerate*FSAMPLES);
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*screen);
	gl_Position = vec4(in_Position,1);
}
void frag() {
	vec3 rp = vec3(0,0.,twave(time)*8.-4.),//camera position
		rd = normalize(vec3(uv.xy+(hash(uvec3(uv.zw,frame)).xy*4.-2.)/screenY,.7)),
		c = vec3(1);
		
	rd.xz *= r2(time*PI*2.);
		
	float l = 0.;
	int i;
	bool flipRefraction = false;
	nsampling = 0.;
	for (i = 0; i < STEPS; i++) {
		float dst = geometry(rp);
		if (dst <= EPS) {
			mat3 mat = material(rp);
			l += mat[1].y;
			c *= mat[0];
			
			rp -= rd*(EPS*2.-dst);
			vec3 nrm = normal(rp);
			mat2x3 tangBinorm = tangentBinormal(nrm);
			
			#if DEBUG
			gl_FragColor = vec4(nrm*.5+.5,1);
			return;
			#endif
			
			vec3 rnd = hash(uvec3(abs(rp)*1000.)+uvec3(frame,frame*3,frame*7));
			float ct = pow(rnd.x,flipRefraction?mat[1].z:mat[1].x),
				st = sqrt(1.-ct*ct),
				pt = rnd.y*PI*2.;
			vec3 surf = normalize(nrm*ct+tangBinorm[0]*(st*cos(pt))+tangBinorm[1]*(st*sin(pt)));
			#if HAS_OPACITY
			if (mat[2].x < rnd.z) {
				rp += rd*(EPS*2.-dst);
				rd = refract(rd,surf,flipRefraction?mat[2].y:1./mat[2].y);
				dst = EPS*REFRACTION_SKIP*(1.+rnd.z);
				flipRefraction = !flipRefraction;
			} else {
			#endif
				rd = reflect(rd,surf);
				dst = 0.;
			#if HAS_OPACITY
			}
			#endif
		}
		rp += rd*dst;
		
		if (length(rp) > 4.) break;
	}
	
	//background + final lighting
	c = min(c,vec3(10.));
	c *= l+background(rp,rd);//*clamp(geometry(rp),0.,1.);
	
	#if ANIMATION
	gl_FragColor = (frame%FSAMPLES==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+vec4(c,1)*1e-2;
	#else
	gl_FragColor = (frame==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+vec4(c,1)*1e-2;
	#endif
}
